#include <iostream>
#include <vector>
#include <random>
#include <thread>
#include <type_traits>
#include <algorithm>
#include <ctime>
#include <map>

constexpr std::uint_fast64_t power_of_2(unsigned int pow)
{
    return 1ull << pow; //return pow == 0 ? 1ull : 2ull * power_of_2(pow-1);
}

template<class int_type>
class bitstats
{
    static_assert(std::numeric_limits<int_type>::is_integer,"int_type must meet numeric_limits<T>::is_integer");
    std::vector<unsigned int> bitcount{sizeof(int_type)* CHAR_BIT/*number of bits*/,0};
    static constexpr std::uint_fast64_t bitvalue[64]={1,2,4,8,16,32,64,128,256,512,1024,2048,4096,8192
                                                      ,power_of_2(14),power_of_2(15),power_of_2(16),power_of_2(17),power_of_2(18),power_of_2(19),power_of_2(20),power_of_2(21)
                                                      ,power_of_2(22),power_of_2(23),power_of_2(24),power_of_2(25),power_of_2(26),power_of_2(27),power_of_2(28),power_of_2(29)
                                                      ,power_of_2(30),power_of_2(31),power_of_2(32),power_of_2(33),power_of_2(34),power_of_2(35),power_of_2(36),power_of_2(37)
                                                      ,power_of_2(38),power_of_2(39),power_of_2(40),power_of_2(41),power_of_2(42),power_of_2(43),power_of_2(44),power_of_2(45)
                                                      ,power_of_2(46),power_of_2(47),power_of_2(48),power_of_2(49),power_of_2(50),power_of_2(51),power_of_2(52),power_of_2(53)
                                                      ,power_of_2(54),power_of_2(55),power_of_2(56),power_of_2(57),power_of_2(58),power_of_2(59),power_of_2(60),power_of_2(61)
                                                      ,power_of_2(62),power_of_2(63)};
public:
    void count(int_type n)
    {
        for(size_t i =0; i < bitcount.size(); ++i)
        {
            if(n & bitvalue[i])
                bitcount[i]++;
        }
    }
    const std::vector<unsigned int>& getBitcount() const{return bitcount;}
};
template<class int_type> constexpr std::uint_fast64_t bitstats<int_type>::bitvalue[64];

template<class RNG, class uint_type = typename RNG::result_type>
std::vector<uint_type> fillRandom(RNG& rng,size_t num)
{
    std::vector<uint_type> vec(num);
    std::generate(vec.begin(),vec.end(),rng);
    return vec;
}

template<class RNG>
class random_test
{
    bitstats<typename RNG::result_type> stats;
    RNG rng;
public:
    random_test(typename RNG::result_type rd = std::time(0)):rng(rd){}
    void run()
    {
        auto vec = fillRandom(rng,5000000);
        for(auto& i: vec )
            stats.count(i);
    }
    const bitstats<typename RNG::result_type>& getStats() const{return stats;}
};

void print_bitcount(const std::vector<unsigned int>& vec)
{
    std::cout << vec.size()<< " bits \n";
    unsigned int i =0;
    for(auto& n:vec)
        std::cout << "bit " << i++ << " " << n << "\n";
    std::cout << "max element: "<< *std::max_element(vec.begin(),vec.end());
    std::cout << "\nmin element: "<< *std::min_element(vec.begin(),vec.end());
    std::cout << "\ndistance: "<<*std::max_element(vec.begin(),vec.end()) - *std::min_element(vec.begin(),vec.end()) << std::endl;
}

int main()
{
    random_test<std::mt19937> mersenne32;
    random_test<std::mt19937_64> mersenne64;
    std::thread t64([&mersenne64](){mersenne64.run();});
    mersenne32.run();
    print_bitcount(mersenne32.getStats().getBitcount());
    t64.join();//*/
    print_bitcount(mersenne64.getStats().getBitcount());
}
